# -*- coding: utf-8 -*-
"""
COPYRIGHT: 
    2019, Agnethe Nedergaard Pedersen and Jonas Wied Pedersen
    2019, VCS Denmark and DTU Environment


LICENSE:
    This file is part of the project "Bellinge - open data"
    
    Licensed under Apache2. 

Purpose
Make simple cleaning of data, converting it to UTC+0 time and ensure that the 
value is in one-minute steps. Apply simple cleaning tools and interpolate when 
missing or faulty data.

input
raw data in a given timeseries. 

returns one-minute continuert timeseries with
- raw value: from raw files
- value_no_errors: raw_value if no errors in the model
Different error marks: 
- ffill: marks if the data value should be like the previous due to savings of data in SCADA system
- stamp: if data is marked with low quality
- man_remove: if dataperiod is marked as error and should be removed
- outbound: marks if the value is out of obvious range given by file
- frozen: marks if data is the same for a given period, and thereby frozen
- outlier: if one point is an outlier compared to the neighbouring values
- frozen_high: if data seems to be frozen at a higher level it is removed. The period of frozen is set to 3 minutes. 
- depth_s: Corrected sensor depth (where grouping for one-minute values and interpolation has occured)
- level: depth_s + zeropoint giving the level in DVR90. 



"""

import os
import numpy as np
import pandas as pd
from bokeh.plotting import figure, output_file, show
from bokeh.models import Span, LinearAxis, Range1d, DatetimeTickFormatter
import bokeh.plotting as bp
import pickle




os.chdir("C:\\_Sager\\_PhD\\Project\\lib")

from Tool_ReadData_2_UTCtime import import_system2000, import_ifix, import_danova, Fill_data_savings_points, NoFill_data_savings_points
from Tools_washErrors import Stamped_error, Out_of_bounds, Outlier_point, Frozen_sensor, Remove_man
from Tool_continousTS_v0_1 import make_continous_series, procces_data



#%% Main

def main(obs_input, man_remove,i):
    
    # LOAD DATA
    name = obs_input.at[i,'IdMeasurement']
    path = obs_input.at[i,'Folderpath'] + obs_input.at[i,'Filename']
    conversion = obs_input.at[i,'Conversion']
    if obs_input.at[i,'Source'] == 'System2000':
        raw_data = import_system2000(path, conversion)#, obs_input.at[i,'StartTime'], obs_input.at[i,'EndTime'] )
        print("System2000 data is loaded:", obs_input.at[i,'IdMeasurement']) 
    elif obs_input.at[i,'Source'] == 'iFix':  
        raw_data = import_ifix(path, conversion)#, obs_input.at[i,'StartTime'], obs_input.at[i,'EndTime'] )
        print("iFix data is loaded:", obs_input.at[i,'IdMeasurement'])
        
    elif obs_input.at[i,'Source'] == 'Danova':  
        raw_data = import_danova(path, conversion)#, obs_input.at[i,'StartTime'], obs_input.at[i,'EndTime'] )
        print("Danova data is loaded:", obs_input.at[i,'IdMeasurement'])       
    else: 
        print("There is no data")

    raw_data.value = raw_data.value.round(3)

    #### FOR TESTING
 #   raw_data = raw_data.iloc[7000:8000,:].reset_index(drop=True)
    ####
    

    #SETTING FLAGS FOR ERRORS
    
    if obs_input.at[i,'Source'] == 'iFix': 
        data = Fill_data_savings_points(raw_data) 
        stamp_id = Stamped_error(data.value, data.quality_ifix, 0) #0 is error in data
    else:
        data = NoFill_data_savings_points(raw_data)
        stamp_id = pd.Series(np.zeros(len(data)),dtype=bool)
    print("Stamp_id and padding of datasaving values is calculated")
    
    man_remove_id = Remove_man(data, man_remove, name)
    
    outbound_id = Out_of_bounds(data.value, obs_input.at[i,'obvious_min'], obs_input.at[i,'obvious_max'])
    print("Outbound_id is calculated")
    
    if name == 'G71F68Y_LevelPS' or obs_input.at[i,'Type'] == 'Position':
        frozen_id2 = pd.Series(np.zeros(len(data.value)),dtype=bool)    
    else: 
        frozen_id2 = Frozen_sensor(data.value, data.time, 0, obs_input.at[i,'window_frozen'])
    print("Frozen_id is calculated")
    
   
    if obs_input.at[i,'Source'] == 'Danova': 
        outlier_id = Outlier_point(data.value, obs_input.at[i,'outlier_threshold'], obs_input.at[i,'outlier_width'])
    else: 
        outlier_id = pd.Series(np.zeros(len(data.value)),dtype=bool)
    print("Outlier_id is calculated")

    
    value_no_errors = data.value.copy()
    value_no_errors[stamp_id | outbound_id | frozen_id2 | man_remove_id | outlier_id] = np.nan  

    # make list with errors instead with numbers
    data_with_tags = pd.DataFrame({'time': data.loc[:,'time'],
                              'raw_value': data.loc[:,'value'],
                              'value_no_errors': value_no_errors,
                              'man_remove': man_remove_id,
                              'ffill': data.loc[:,'ffill'],
                              'stamp': stamp_id,
                              'outbound': outbound_id,
                              'frozen': frozen_id2,
                              'outlier': outlier_id})
        
    print("Proccessed data made")
    

    ts, ts_meanminmax = make_continous_series(data_with_tags, obs_input.at[i,'StartTime'], obs_input.at[i,'EndTime'])
    
    ts['level'] = procces_data(ts.depth_s, obs_input, obs_input.at[i,'IdMeasurement'], obs_input.at[i,'Source'], obs_input.at[i,'Version'])
   
    return ts, ts_meanminmax, data_with_tags, raw_data
 
def Plot_datawash(output, data_with_tags, i, raw_values, outfile):
    ####################################        
    data = data_with_tags.copy()
    
    fstamp = data.loc[data['stamp'],:]
    foutbound = data.loc[data['outbound'],:]
    ffrozen = data.loc[data['frozen'],:]
    fmanremove = data.loc[data['man_remove'],:]
    foutlier = data.loc[data['outlier'],:]

    
    name = obs_input.at[i, 'IdMeasurement']
    output_file(outfile, title = name) 
    
    
    p = bp.figure(plot_width=1500, plot_height=800, 
                x_axis_type="datetime", 
                title=name) 
    
    p.xaxis.formatter = DatetimeTickFormatter(
            years = "%d-%m-%Y",
            days="%d-%m-%Y %H:%M",
            months="%d-%m-%Y %H:%M",
            hours="%d-%m-%Y %H:%M",
            minutes="%d-%m-%Y %H:%M")
    
    p.line(x=raw_values['time'], y=raw_values['value'], legend_label = 'raw value', line_width = 1, color='lightblue')
    
    p.circle(x=data['time'], y=data['raw_value'], legend_label = name, size =4, color='grey')
    p.line(x=output['time'], y=output['depth_s'], legend_label = name, line_width = 1, color='black')
    
    p.circle(x=fstamp['time'], y=fstamp['raw_value'], legend_label = 'stamp', size = 2, color = 'blue')
    p.circle(x=foutbound['time'], y=foutbound['raw_value'], legend_label = 'outbound', size = 2, color = 'red')
    p.circle(x=ffrozen['time'], y=ffrozen['raw_value'], legend_label = 'frozen', size = 4, color = 'green')
    p.circle(x=fmanremove['time'], y=fmanremove['raw_value'], legend_label = 'manual remove', size = 4, color = 'pink')
    p.circle(x=foutlier['time'], y=foutlier['raw_value'], legend_label = 'single outlier', size = 4, color = 'violet')
    
    show(p)   
    return

def extract_dates(df, time, start, stop):
    df[time] = pd.to_datetime(df[time], format="%Y-%m-%d %H:%M:%S")
    model_ex = df[df[time].between(pd.to_datetime(start), pd.to_datetime(stop))]

    return model_ex

#%% Run program

os.chdir("C:\\_Sager\\_PhD\\Project")

if not os.path.exists('./observations/proc/'):
    os.makedirs('./observations/proc/')
    
if not os.path.exists('./output/proc_data/'):
    os.makedirs('./output/proc_data/')    

obs_input= pd.read_csv('./etc/obs_input.csv', sep=";")
man_remove = pd.read_csv('./etc/manual_remove.csv', sep=";")
    
Area = 'Bellinge'

for i in range(0,len(obs_input)): 
    if obs_input.at[i,'Area'] == Area: 
        if obs_input.at[i, 'Type'] != 'Power': 
            name = obs_input.at[i,'IdMeasurement']
            output, ts_meanminmax, data_with_tags, raw_data = main(obs_input, man_remove,i)#, output)
            
            
            save_file="./observations/proc/"+obs_input.at[i,'IdMeasurement']+"_"+obs_input.at[i,'Source']+obs_input.at[i,'Version']
            
            output.to_csv(save_file+"_proc_v6.csv", index = True, header=True)
         
            ts_obs1 = output[['time', 'level']]
            with open("./observations/proc/"+name+"_"+obs_input.at[i,'Source']+obs_input.at[i,'Version']+"_obs.p", 'wb') as f: 
                pickle.dump(ts_obs1, f)
            
            
            print("finished with {}".format(name))           
            outfile = "./output/proc_data/"+name+"_"+str(i)+"_v6.html"
            Plot_datawash(output, data_with_tags,i, raw_data, outfile)

###########
# Combine to one timeseries
##############

listofids = list(obs_input['IdMeasurement'])

for i in range(0, len(obs_input)):
    if obs_input.at[i,'Area'] == Area: 
        if obs_input.at[i, 'Type'] != 'Power':# and obs_input.at[i,'Area'] == Area:
            name = obs_input.at[i,'IdMeasurement']
            idx = [i for i in range(len(listofids)) if listofids[i] == name] 
            if i == idx[-1]: #only run the rest if it is the last unique id.
                total = pd.DataFrame()
                for j in range(0,len(idx)):
                    with open("./observations/proc/"+name+"_"+obs_input.at[idx[j],'Source']+obs_input.at[idx[j],'Version']+"_obs.p", 'rb') as f: 
                        temp = pickle.load(f)
                    total = total.append(temp, ignore_index= True)
                total.sort_values(by='time')
        
                with open("./observations/proc/"+name+"_obs.p", 'wb') as f: 
                    pickle.dump(total, f)